import threading
import time
import logging
from gql import gql
from gql.transport.exceptions import TransportServerError
from datetime import datetime
from dateutil.parser import isoparse

from .data import Issue
from .data import PullRequest
from .data import Repository


def _make_issue_kwargs(gh_data):
    repo = Repository(
        owner=gh_data["repository"]["owner"]["login"],
        name=gh_data["repository"]["name"],
    )
    if gh_data["author"] is None:
        # https://github.com/ghost
        author = "ghost"
    else:
        author = gh_data["author"]["login"]
    return {
        "repo": repo,
        "author": author,
        "created_at": isoparse(gh_data["createdAt"]),
        "updated_at": isoparse(gh_data["updatedAt"]),
        "number": int(gh_data["number"]),
        "title": gh_data["title"],
        "url": gh_data["url"],
    }


def _make_pr(gh_data):
    approved = False
    changes_requested = False
    for hc in gh_data["hovercard"]["contexts"]:
        if "reviewDecision" in hc:
            if hc["reviewDecision"] == "CHANGES_REQUESTED":
                changes_requested = True
            elif hc["reviewDecision"] == "APPROVED":
                approved = True
            break
    return PullRequest(
        approved=approved,
        changes_requested=changes_requested,
        draft=bool(gh_data["isDraft"]),
        **_make_issue_kwargs(gh_data),
    )


def _make_issue(gh_data):
    return Issue(**_make_issue_kwargs(gh_data))


FRAGMENT_ISSUE = """
fragment issueFields on Issue {
    author {
        login
    }
    createdAt
    number
    title
    updatedAt
    url
    repository {
        name
        owner {
            login
        }
    }
}
"""
FRAGMENT_PR = """
fragment prFields on PullRequest {
    author {
        login
    }
    createdAt
    number
    title
    updatedAt
    url
    isDraft
    hovercard {
        contexts {
            ... on ReviewStatusHovercardContext {
                reviewDecision
            }
        }
    }
    repository {
        name
        owner {
            login
        }
    }
}
"""


class IssueQuery:

    def __init__(self, *, first=100, after=None, states=("OPEN",)):
        self.first = first
        self.after = after
        self.states = states

    def __str__(self):
        parts = [f"issues(first: {self.first}"]
        if self.after:
            parts.append(f', after: "{self.after}"')
        if self.states:
            parts.append(", states: [")
            parts.append(",".join(self.states))
            parts.append("]")
        parts.append(
            ") { nodes { ...issueFields } pageInfo { endCursor hasNextPage } }"
        )
        return "".join(parts)


class PRQuery:

    def __init__(self, *, first=100, after=None, states=("OPEN",)):
        self.first = first
        self.after = after
        self.states = states

    def __str__(self):
        parts = [f"pullRequests(first: {self.first}"]
        if self.after:
            parts.append(f', after: "{self.after}"')
        if self.states:
            parts.append(", states: [")
            parts.append(",".join(self.states))
            parts.append("]")
        parts.append(") { nodes { ...prFields } pageInfo { endCursor hasNextPage } }")
        return "".join(parts)


class IssueLoader:

    def __init__(self, gql_client, repos, cache, progress_callback):
        self._client = gql_client
        self._repos = tuple(repos)
        self._cache = cache
        self._lock = threading.Lock()
        self._logger = logging.getLogger("IssueLoader")
        self._thread = threading.Thread(daemon=True, target=self._run)
        self._progress_callback = progress_callback
        self._thread.start()
        self._stop = False

    def stop(self):
        self._stop = True

    def _run(self):
        try:
            self._load_all_issues(
                repos_per_query=5, progress_callback=self._progress_callback
            )
        except:
            self._logger.exception("Exception in IssueLoader thread")
        time.sleep(15)
        while not self._stop:
            try:
                # Uses search API to get updated issues and PRs
                self._update_all_issues()
            except:
                self._logger.exception("Exception in IssueLoader thread")
            time.sleep(15)
        self._logger.debug("IssueLoader thread exiting")

    def _load_all_issues(self, repos_per_query, progress_callback=None):
        num_repos_at_start = len(self._repos)

        # These dicts indicate if repos have more issues or PRs to query
        issue_page_info = {}
        pr_page_info = {}
        for r in self._repos:
            # Use "None" to mean we haven't queried anything yet
            issue_page_info[r] = {"endCursor": None}
            pr_page_info[r] = {"endCursor": None}

        # Outer loop runs until it finishes exploring all issues and PRs
        # on all repos
        next_queries = []
        issue_count = 0
        pr_count = 0
        # Create a session to avoid closing and reopening the connection
        # for each query
        session = self._client.connect_sync()
        try:
            while issue_page_info or pr_page_info:
                # This loop looks for repos that still need to be explored
                for r in self._repos:
                    issue_query = None
                    pr_query = None
                    if r in issue_page_info:
                        ipi = issue_page_info[r]
                        issue_query = IssueQuery(after=ipi["endCursor"])
                    if r in pr_page_info:
                        prpi = pr_page_info[r]
                        pr_query = PRQuery(after=prpi["endCursor"])

                    if issue_query or pr_query:
                        repo_query = f'r{id(r)}: repository(owner: "{r.owner}", name: "{r.name}")'
                        repo_query += "{"
                        if issue_query:
                            repo_query += str(issue_query)
                        if pr_query:
                            repo_query += str(pr_query)
                        repo_query += "}"
                        next_queries.append(repo_query)

                    if len(next_queries) == repos_per_query:
                        # Found enough repos, ditch the for loop
                        break

                # Execute the next query
                joined_queries = "\n".join(next_queries)
                query_str = f"""
                    query {{
                        {joined_queries}
                    }}
                    """
                next_queries = []
                # It's an error to include unused fragments,
                # so only include a fragment if it's used.
                if issue_page_info:
                    query_str += FRAGMENT_ISSUE
                if pr_page_info:
                    query_str += FRAGMENT_PR
                query = gql(query_str)

                # Execute with exponential backoff in case we hit secondary
                # rate limits from Github.
                executed = False
                error_count = 0
                while not executed:
                    try:
                        result = session.execute(query)
                        executed = True
                    except TransportServerError:
                        self._logger.exception("Error when loading initial issues")
                        error_count += 1
                        time.sleep(15 + 4**error_count)

                # Figure out what repos the query included,
                # and add the issues and PRs to the cache
                for r in self._repos:
                    key = f"r{id(r)}"
                    if key not in result:
                        # Query didn't include this repo
                        continue
                    repo_result = result[key]
                    if "issues" in repo_result:
                        for issue in repo_result["issues"]["nodes"]:
                            self._cache.insert(_make_issue(issue))
                            issue_count += 1
                        if repo_result["issues"]["pageInfo"]["hasNextPage"]:
                            issue_page_info[r] = repo_result["issues"]["pageInfo"]
                        else:
                            del issue_page_info[r]
                    if "pullRequests" in repo_result:
                        for pr in repo_result["pullRequests"]["nodes"]:
                            self._cache.insert(_make_pr(pr))
                            pr_count += 1
                        if repo_result["pullRequests"]["pageInfo"]["hasNextPage"]:
                            pr_page_info[r] = repo_result["pullRequests"]["pageInfo"]
                        else:
                            del pr_page_info[r]
                if progress_callback:
                    i_max = pr_max = num_repos_at_start
                    progress_callback(
                        (i_max - len(issue_page_info) + pr_max - len(pr_page_info))
                        / (i_max + pr_max)
                    )
            self._logger.info(f"Loaded {issue_count} issues and {pr_count} PRs")
        finally:
            self._client.close_sync()

    def _update_all_issues(self):

        updated_time = self._cache.newest_update_time().isoformat()

        def make_query(extra):
            gh_search = f"{extra} is:open updated:>{updated_time} "
            gh_search += " ".join([f"repo:{r.owner}/{r.name}" for r in self._repos])
            query_parts = ["{"]
            query_parts.append(f'search(first: 1, query: "{gh_search}", type: ISSUE) ')
            query_parts.append("{ nodes {...issueFields ...prFields} }")
            query_parts.append("}")
            query_parts.append(FRAGMENT_ISSUE)
            query_parts.append(FRAGMENT_PR)
            return " ".join(query_parts)

        # TODO use pagination for unlikely case of more than 100 updated issues
        # Must query for issues and PRs separately
        # https://github.com/orgs/community/discussions/149046
        issues = self._client.execute(gql(make_query("is:issue")))
        for i in issues["search"]["nodes"]:
            self._cache.insert(_make_issue(i))

        prs = self._client.execute(gql(make_query("is:pr")))
        for pr in prs["search"]["nodes"]:
            self._cache.insert(_make_pr(pr))



from dataclasses import dataclass
from datetime import datetime


@dataclass(frozen=True)
class Repository:
    owner: str = ""
    name: str = ""


@dataclass
class Issue:
    repo: Repository = Repository
    author: str = ""
    created_at: datetime = None
    updated_at: datetime = None
    number: int = 0
    title: str = ""
    url: str = ""
    is_pr: bool = False


@dataclass
class PullRequest(Issue):
    is_pr: bool = True
    approved: bool = False
    changes_requested: bool = False
    draft: bool = False


def is_same_issue(l, r):
    return l.repo == r.repo and l.number == r.number



import treadi.config as tc

import pytest
from unittest.mock import patch


def test_config_dir():
    assert tc.config_dir().name == "TreadI"


def test_initialize_config(tmp_path):
    tc._initialize_config_dir(tmp_path)

    assert (tmp_path / "TreadI.config.version.txt").exists()
    assert (tmp_path / "TreadI.config.version.txt").is_file()
    assert (tmp_path / "repository_lists.d").exists()
    assert (tmp_path / "repository_lists.d").is_dir()
    assert (tmp_path / "repository_lists.d" / "README.md").exists()
    assert (tmp_path / "repository_lists.d" / "README.md").is_file()
    assert (tmp_path / "repository_lists.d" / "25_Public_repos_I_own.txt").exists()
    assert (tmp_path / "repository_lists.d" / "25_Public_repos_I_own.txt").is_file()

    content = (
        tmp_path / "repository_lists.d" / "25_Public_repos_I_own.txt"
    ).read_text()
    assert "@me" in content


def test_config_repo_list_names(tmp_path):
    tc._initialize_config_dir(tmp_path)
    with patch("treadi.config.config_dir", lambda: tmp_path):
        c = tc.Config()
        assert "Public repos I own" in c.repository_list_names()


def test_config_repo_list(tmp_path):
    tc._initialize_config_dir(tmp_path)
    with patch("treadi.config.config_dir", lambda: tmp_path):
        c = tc.Config()
        repo_list = c.repository_list("Public repos I own")
        assert repo_list.rank == 25
        assert repo_list.display_name == "Public repos I own"
        assert "@me" in repo_list.content



from kivy.clock import Clock
from kivy.uix.screenmanager import Screen

from .issue_loading_screen import IssueLoadingScreen


class RepoLoadingScreen(Screen):

    def __init__(self, loader, **kwargs):
        self._loader = loader
        if "name" not in kwargs:
            kwargs["name"] = "repo-loading-screen"
        super().__init__(**kwargs)

    def on_pre_enter(self):
        self._loader.begin_loading(self.switch_to_issue_loading)

    def switch_to_issue_loading(self, repos):

        def _switch(dt):
            self.manager.switch_to(IssueLoadingScreen(repos), direction="left")

        Clock.schedule_once(lambda dt: self._loader.cleanup())
        Clock.schedule_once(_switch)



import webbrowser

from kivy.clock import Clock
from kivy.properties import ObjectProperty
from kivy.uix.screenmanager import Screen

from .. import auth


class LoginScreen(Screen):

    device_flow = ObjectProperty(
        auth.DeviceFlow(
            device_code="",
            user_code="",
            verification_uri="",
            interval=0,
            expires_in=0,
        ),
        rebind=True,
    )
    token_response = ObjectProperty()

    def on_enter(self):
        self.start_device_flow()

    def start_device_flow(self):
        self.device_flow = auth.start_device_flow()
        print("Device flow started: ", self.device_flow)
        Clock.schedule_once(lambda dt: self.check_auth(), self.device_flow.interval)

    def open_browser(self, url):
        webbrowser.open(url)

    def check_auth(self):
        response = auth.ask_for_token(self.device_flow)
        match response.status:
            case auth.Status.AUTHORIZATION_PENDING:
                Clock.schedule_once(
                    lambda dt: self.check_auth(), self.device_flow.interval
                )
            case auth.Status.EXPIRED_TOKEN:
                self.start_device_flow()
            case auth.Status.ACCESS_DENIED:
                raise RuntimeError(
                    "TODO nice error message when user denies TreadI App"
                )
            case auth.Status.ACCESS_GRANTED:
                # Listeners on the token_response property are notified here
                self.token_response = response
            case _:
                raise RuntimeError(
                    f"TODO nice error message when other error encountered{response}"
                )



from kivy.app import App
from kivy.uix.button import Button
from kivy.uix.screenmanager import Screen

from .. import config
from ..repo_loader import parse_repository_list
from .repo_loading_screen import RepoLoadingScreen


class RepoButton(Button):
    """RepoPickerScreen button.

    This is a class so I can customize in kv language.
    """

    def __init__(self, repo_picker, *args, **kwargs):
        self._repo_picker = repo_picker
        self._display_name = kwargs["text"]
        super().__init__(*args, **kwargs)

    def on_release(self):
        self._repo_picker.use_repo_list(self._display_name)


class RepoPickerScreen(Screen):

    def __init__(self, *args, **kwargs):
        self._config = config.Config()
        if "name" not in kwargs:
            kwargs["name"] = "repo-picker-screen"
        super().__init__(*args, **kwargs)

    def on_pre_enter(self):
        for name in self._config.repository_list_names():
            btn = RepoButton(repo_picker=self, text=name)
            self.ids.stack.add_widget(btn)

    def use_repo_list(self, display_name):
        repo_list = self._config.repository_list(display_name)
        repo_loader = parse_repository_list(
            repo_list.content,
            # TODO pass this into repo loader
            gql_client=App.get_running_app().gql_client,
        )
        self.manager.switch_to(RepoLoadingScreen(repo_loader), direction="left")
        App.get_running_app().title += f": {display_name}"



from pathlib import Path

from platformdirs import user_config_dir

from dataclasses import dataclass
from datetime import datetime

import logging
import re


@dataclass(frozen=True)
class RepositoryList:
    path: Path
    display_name: str
    rank: int
    content: str


def config_dir():
    """Return the root path to TreadI's config directory."""
    return Path(user_config_dir("TreadI"))


def create_or_update_config():
    """Create an initialize the config directory if none exists."""
    cdir = config_dir()
    if cdir.exists():
        # TODO Upgrade config if necessary
        return

    # Create config dir
    cdir.mkdir()
    _initialize_config_dir(cdir)


_REPO_LIST_DIR = "repository_lists.d"
_VERSION_FILE = "TreadI.config.version.txt"
# Increment this any time a built-in config changes
# This enables updating the configs
# Format: "vYYYY_MM_DD.12345"
#   The date gives an idea of how recent information is.
#   If for some reason I need to bump the config version twice in
#   one day, the `.number` at the end lets me increment it again.
#   Most config versions should end in `.0`
_CONFIG_VERSION = "v2025-03-29.0"
# Matches file names of repository list files.
_REPO_LIST_REGEX = re.compile("(?P<rank>[0-9][0-9])_(?P<name>[-a-zA-Z0-9._]+).txt")


def _initialize_config_dir(cdir: Path):
    """Initialize the config directory.

    This expects a directory as an argument, and populates the
    directory with default configuration.
    """
    assert cdir.exists()
    assert cdir.is_dir()

    # Make a version file
    (cdir / _VERSION_FILE).write_text(_CONFIG_VERSION)

    # Create a repository lists directory
    external_repos_d = cdir / _REPO_LIST_DIR
    external_repos_d.mkdir()

    # Copy repository lists into the repository lists directory
    internal_repos_d = Path(__file__).parent.resolve() / _REPO_LIST_DIR
    for internal_file in internal_repos_d.iterdir():
        content = internal_file.read_text()
        header = None
        if internal_file.suffix == ".txt":
            header = (
                "#########\n"
                + "# Built-in repository list\n"
                + "# Any edits may be overwritten the next time you upgrade TreadI\n"
                + f"# TreadI Config version: {_CONFIG_VERSION}\n"
                + "#########\n"
            )
        external_file = external_repos_d / internal_file.name
        if header:
            content = "\n".join([header, content])
        external_file.write_text(content)


def _make_repo_list_display_name(raw_name):
    return raw_name.replace("_", " ").strip()


class Config:

    def __init__(self):
        self._logger = logging.getLogger("Config")

    def repository_list_names(self):
        """Return a list of repository list display names."""
        repo_list_d = config_dir() / _REPO_LIST_DIR
        rank_name = []
        for file in repo_list_d.iterdir():
            if file.name == "README.md":
                continue
            match = _REPO_LIST_REGEX.match(file.name)
            if match is None:
                self._logger.warning(f"Invalid repository list file name: {file.name}")
                continue
            rank_name.append(
                (
                    int(match.group("rank")),
                    _make_repo_list_display_name(match.group("name")),
                )
            )

        # Rely on stable sorting to make rank primary key, and display name secondary
        rank_name.sort(key=lambda i: i[1].lower())
        rank_name.sort(key=lambda i: i[0])
        return [i[1] for i in rank_name]

    def repository_list(self, display_name):
        """Given a display name, return the repository list content.

        raise RuntimeError if a repository list is not found.
        """
        repo_list_d = config_dir() / _REPO_LIST_DIR
        for file in repo_list_d.iterdir():
            match = _REPO_LIST_REGEX.match(file.name)
            if match is None:
                self._logger.warning(f"Invalid repository list file name: {file.name}")
                continue
            found_name = _make_repo_list_display_name(match.group("name"))
            if found_name == display_name:
                return RepositoryList(
                    path=file,
                    display_name=display_name,
                    rank=int(match.group("rank")),
                    content=file.read_text(),
                )
        raise RuntimeError("No such repository list: ", display_name)



import kivy

kivy.require("2.3.1")

from kivy.app import App
from kivy.config import Config as KvConfig

KvConfig.set("graphics", "resizable", False)
KvConfig.set("input", "mouse", "mouse,disable_multitouch")
KvConfig.set("kivy", "exit_on_escape", 0)

from kivy.core.window import Window

Window.size = (500, 558)

from kivy.uix.popup import Popup
from kivy.uix.screenmanager import ScreenManager

import logging
import pathlib
import sys

from gql import gql
from gql import Client
from gql.transport.exceptions import TransportServerError
from gql.transport.requests import RequestsHTTPTransport

from . import auth
from . import config
from .issue_cache import IssueCache
from .screens.issue_screen import IssueScreen
from .screens.login_screen import LoginScreen
from .screens.repo_picker_screen import RepoPickerScreen


def make_gql_client(access_token):
    transport = RequestsHTTPTransport(
        url="https://api.github.com/graphql",
        headers={
            "Authorization": f"bearer {access_token}",
        },
        verify=True,
        retries=3,
    )
    schema_path = pathlib.Path(__file__).parent.resolve() / "schema.docs.graphql"
    return Client(transport=transport, schema=schema_path.read_text())


def check_gql_works(gql_client):
    query = gql(
        """
        query {
            viewer {
                login
            }
        }
        """
    )
    try:
        gql_client.execute(query)
        return True
    except TransportServerError:
        return False


class InvalidTokenPopup(Popup):
    pass


class TreadIApp(App):

    # Globals accessible to all screens
    gql_client = None
    issue_loader = None
    issue_cache = IssueCache()
    sm = None

    def make_client_from_response(self, token_response):
        if token_response.status == auth.Status.ACCESS_GRANTED:
            self.gql_client = make_gql_client(token_response.access_token)
            return check_gql_works(self.gql_client)
        return False

    def on_login_result(self, _, token_response):
        if self.make_client_from_response(token_response):
            auth.store_refresh_token(token_response.refresh_token)
            self.switch_to_pick_repos()
        else:
            raise RuntimeError(
                f"TODO more graceful response to login failure {token_response}"
            )

    def switch_to_pick_repos(self, direction="left"):
        """Called to switch to repo picker screen and reset."""
        # Reset title away from repo list name
        self.title = "TreadI"
        # This avoids a circular dependency in screen modules
        if self.issue_loader is not None:
            self.issue_loader.stop()
        self.issue_loader = None
        self.issue_cache = IssueCache()
        self.sm.transition.direction = direction
        self.sm.switch_to(RepoPickerScreen())

    def build(self):
        self.icon = "treadi_icon.png"
        self.sm = ScreenManager()
        config.create_or_update_config()

        # Prefer a PAT because a user might have granted it
        # private repo access and want to maintain a private repo.
        # The device flow based login can only do public repos.
        pat = auth.get_personal_access_token()
        if pat.status == auth.Status.ACCESS_GRANTED:
            if not self.make_client_from_response(pat):
                p = InvalidTokenPopup()
                p.bind(on_dismiss=sys.exit)
                p.open()
            else:
                self.switch_to_pick_repos()
            return self.sm

        token_response = auth.cycle_cached_token()
        if not self.make_client_from_response(token_response):
            # Ask user to login
            login_screen = LoginScreen()
            login_screen.bind(token_response=self.on_login_result)
            self.sm.switch_to(login_screen)
        else:
            self.switch_to_pick_repos()

        return self.sm


def main():
    TreadIApp().run()


if __name__ == "__main__":
    main()
